/* this is session.c file.
 * this is part of the libspopc library sources
 * copyright  2002 Benoit Rouits <brouits@free.fr>
 * released under the terms of GNU LGPL
 * (GNU Lesser General Public Licence).
 * libspopc offers simple API for a pop3 client (MTA).
 * See RFC 1725 for pop3 specifications.
 * more information on http://brouits.free.fr/libspopc/
 */

/*I casted 'port' to true unsigned, now, tell me if warning continues */
/*; #pragma warning(disable: 4761)*/ /* disable "integral size mismatch in argument" -warning */

#ifdef WIN32
#include <winsock.h>
#include <io.h>
#else
#include <netdb.h>
#include <unistd.h>
#include <netinet/in.h>
#include <sys/socket.h>
#endif


#include <stdlib.h>
#include <string.h> /* strdup */
#include <stdio.h> /* perror */
#include <sys/types.h>
#define socklen_t int /* actually true on most systems */
#include "libspopc.h"

extern int pop3_recv (pop3sock_t sock, char* buf, int len);

#ifdef USE_SSL

static int SSL_UP = 0;

static char* SSL_CERT = NULL;


void pop3_cert_setup(const char *certfile)
{
	if (SSL_CERT) free(SSL_CERT);
	SSL_CERT=NULL;
	if (certfile) SSL_CERT=strdup(certfile);
}


int ssl_verify_callback(int ok, X509_STORE_CTX *ctx) {
    if (!ok) {
        int err;
        err = X509_STORE_CTX_get_error(ctx);
        return 0;
    }
    return 1;
}



pop3sock_t ssl_prepare(const int port){
pop3sock_t sock;
	sock =(pop3sock_t)malloc(sizeof(pop3sock));
	sock->sock=socket(AF_INET,SOCK_STREAM,0);
	if(-1==sock->sock){
		perror("ssl_prepare.socket");
                free(sock);
		return(NULL);
	}
	if ( 995 == port ) {
		if (0 == SSL_UP) SSL_UP=SSL_library_init();

		if (1 != SSL_UP) {
			close(sock->sock);
			free(sock);
			perror("Error initializing SSL");
			return(NULL);
		}
		SSL_load_error_strings();
  	    	sock->ctx = SSL_CTX_new(SSLv23_client_method());
		if (NULL == sock->ctx) {
			close(sock->sock);
			free(sock);
			perror("SSL v23 not available");
			return(NULL);
		}   
		if ( SSL_CERT != NULL ) {
			SSL_CTX_load_verify_locations(sock->ctx, SSL_CERT, 0);
			SSL_CTX_set_verify(sock->ctx, SSL_VERIFY_PEER, &ssl_verify_callback);
		}
		sock->ssl = SSL_new(sock->ctx);
		if (NULL == sock->ssl) {
			close(sock->sock);
                        SSL_CTX_free(sock->ctx);
			free(sock);
			perror("Error creating SSL context");
			return(NULL);
		}
		SSL_set_fd(sock->ssl, sock->sock);
	} else {
		sock->ssl=NULL;
		sock->ctx=NULL;
	}
	return sock;
}

#endif

pop3sock_t pop3_prepare(const char* servername, const int port, struct sockaddr_in* connection, struct hostent* server){
/* prepares the pop session and returns a socket descriptor */
pop3sock_t sock;
struct hostent* hostent_buf;
int i;
#ifdef WIN32

	WSADATA wsaData;

	if (WSAStartup(MAKEWORD(2, 0), &wsaData) != 0)
	{
	exit(1);
	}

#endif

	memset((char*)connection,0,sizeof(struct sockaddr_in));

	hostent_buf=gethostbyname(servername);
	if(!hostent_buf){
		herror("pop3_prepare.gethostbyname");
		return BAD_SOCK;
	}
    /* begin hostent deep copy */
    memmove(server,hostent_buf,sizeof(struct hostent));
#if 1
    server->h_name=strdup(hostent_buf->h_name);
    server->h_aliases=NULL;
    i=0;
    while (hostent_buf->h_aliases[i])
    {
        server->h_aliases=realloc(server->h_aliases,(i+1)*sizeof(char*));
        if(hostent_buf->h_aliases[i])
            server->h_aliases[i]=strdup(hostent_buf->h_aliases[i]);
        else server->h_aliases[i]=NULL;
        i++;
    }
    server->h_addr_list=NULL;
    i=0;
    while (hostent_buf->h_addr_list[i])
    { /* has at least one adress */
        server->h_addr_list=realloc(server->h_addr_list,(i+1)*sizeof(char*));
        server->h_addr_list[i]=malloc(server->h_length);
        memmove(server->h_addr_list[i],hostent_buf->h_addr_list[i],server->h_length);
        i++;
    } 
    server->h_addr_list=realloc(server->h_addr_list, (i+1)*sizeof(char*));
/*
    server->h_addr_list[i]=malloc(1);
    server->h_addr_list[i]='\0';
*/
    server->h_addr_list[i]=NULL;
    /* end hostent deep copy */
#endif
	memmove((char*)&(connection->sin_addr.s_addr),server->h_addr,server->h_length);
	connection->sin_family=AF_INET;
	connection->sin_port=htons(port?(unsigned short int)port:(unsigned short int)110);	/* integral size mismatch in argument - htons(port)*/

#ifdef USE_SSL
	sock = ssl_prepare(port);
#else
	sock=socket(AF_INET,SOCK_STREAM,0);
	if(BAD_SOCK==sock){
		perror("pop3_prepare.socket");
	}
#endif
	return(sock);
}

char* pop3_connect(pop3sock_t sock, struct sockaddr_in* connection){
/* connects to the server through the sock and returns server's welcome */
int r;
char* buf;


#ifdef USE_SSL
	r=connect(sock->sock,(struct sockaddr*)connection,(socklen_t)sizeof(*connection));
	if ( r!=-1  && sock->ssl ) {
		if (SSL_connect(sock->ssl) == -1) {
			close(sock->sock);
			return(NULL);
		}
		if ( SSL_CERT ) {
			if ( !SSL_get_peer_certificate(sock->ssl) ) {
				close(sock->sock);               
				return(NULL);
	                }
		}
	}   
#else
	r=connect(sock,(struct sockaddr*)connection,(socklen_t)sizeof(*connection));
#endif

	if(r==-1){
		perror("pop3_connect.connect");
		return(NULL);
	}


	buf=(char*)malloc(POPBUF);
	if(!buf){
		perror("pop3_connect.malloc");
		return(NULL);
	}
	r=pop3_recv(sock,buf,POPBUF); /* 512 is enough, as say RFC 1939 */
	buf[r]='\0';

	return(buf);
}

void pop3_disconnect(pop3sock_t sock, struct hostent* server){
    int i;
/* close socket  */
#ifdef USE_SSL
	if (sock->sock>0) 
    {
        shutdown(sock->sock,SHUT_RDWR);
        close(sock->sock);
    }
    if (NULL != sock->ssl) {
/*        SSL_shutdown(sock->ssl);*/
		SSL_free(sock->ssl);
	}
	if ( NULL != sock->ctx ) {
		SSL_CTX_free(sock->ctx);
	}
	
	free(sock);
#else
	if (sock>0) {
        shutdown(sock,SHUT_RDWR);
        close(sock);
    }
#endif
    /* free deep copy of struct hostent */
    i=0;
    while(NULL != server->h_addr_list[i]) {
    /* has at least one adress */
        free(server->h_addr_list[i]);
        i++;
    }
/*    free(server->h_addr_list[i]); FIXME  */
    free(server->h_addr_list);
    free(server->h_name);
#ifdef WIN32
	WSACleanup();
#endif
}
